function W5()
    % Create System objects used for reading video, detecting moving objects, and displaying the results.
    obj = setupSystemObjects();
    tracks = initializeTracks(); 
    nextId = 1; 
    videoWriter = vision.VideoFileWriter('Basketball_First_Roll.mp4','FileFormat','MPEG4');
    figure; imshow('Basketball.jpg'); title('Direction of Detection');

    % Detect moving objects, and track them across video frames.
    while hasFrame(obj.reader)
        frame = readFrame(obj.reader);
        [centroids, bboxes, mask] = detectObjects(frame);
        predictNewLocationsOfTracks();
        [assignments, unassignedTracks, unassignedDetections] = ...
            detectionToTrackAssignment();

        updateAssignedTracks();
        updateUnassignedTracks();
        deleteLostTracks();
        createNewTracks();

        displayTrackingResults();
    end

    release(videoWriter); % Release the video writer object once done

%% Create System Objects
function obj = setupSystemObjects()
    obj.reader = VideoReader('Basketball_First_Roll_3.mp4');
    
    % Create two video players, one to display the video,
    % and one to display the foreground mask.
    % Scaling factor for the video display
    scalingFactor = 0.5; % Adjust this to scale down the video size

    % Calculate the new dimensions after scaling
    scaledWidth = round(3840 * scalingFactor);
    scaledHeight = round(2160 * scalingFactor);

    % Set the position of the video players with scaled dimensions
    obj.videoPlayer = vision.VideoPlayer('Position', [20, 400, scaledWidth, scaledHeight]);
    obj.maskPlayer = vision.VideoPlayer('Position', [scaledWidth + 40, 400, scaledWidth, scaledHeight]);
 
    % Create System objects for foreground detection and blob analysis
    obj.detector = vision.ForegroundDetector('NumGaussians', 3, ...
        'NumTrainingFrames', 50, 'MinimumBackgroundRatio', 0.7);
    obj.blobAnalyser = vision.BlobAnalysis('BoundingBoxOutputPort', true, ...
        'AreaOutputPort', false, 'CentroidOutputPort', true, ...
        'MinimumBlobArea', 250);
end

%% Initialize Tracks
function tracks = initializeTracks()
    % Create an empty array of tracks.
    tracks = struct(...
        'id', {}, ...
        'bbox', {}, ...
        'kalmanFilter', {}, ...
        'age', {}, ...
        'totalVisibleCount', {}, ...
        'consecutiveInvisibleCount', {});
end

%% Detect Objects
function [centroids, bboxes, mask] = detectObjects(frame)
    mask = obj.detector.step(frame);
    mask = imopen(mask, strel('rectangle', [3, 3]));
    mask = imclose(mask, strel('rectangle', [15, 15])); 
    mask = imfill(mask, 'holes');

    [centroids, bboxes] = obj.blobAnalyser.step(mask);
end

%% Predict New Locations of Existing Tracks
function predictNewLocationsOfTracks()
    for i = 1:length(tracks)
        bbox = tracks(i).bbox;
        
        % Predict the current location of the track.
        predictedCentroid = predict(tracks(i).kalmanFilter);
        
        % Shift the bounding box so that its center is at the predicted location.
        predictedCentroid = int32(predictedCentroid) - bbox(3:4) / 2;
        tracks(i).bbox = [predictedCentroid, bbox(3:4)];
    end
end

%% Assign Detections to Tracks
function [assignments, unassignedTracks, unassignedDetections] = detectionToTrackAssignment()
    nTracks = length(tracks);
    nDetections = size(centroids, 1);
    
    % Compute the cost of assigning each detection to each track.
    cost = zeros(nTracks, nDetections);
    for i = 1:nTracks
        cost(i, :) = distance(tracks(i).kalmanFilter, centroids);
    end
    
    % Solve the assignment problem.
    costOfNonAssignment = 20;
    [assignments, unassignedTracks, unassignedDetections] = ...
        assignDetectionsToTracks(cost, costOfNonAssignment);
end

%% Update Assigned Tracks
function updateAssignedTracks()
    numAssignedTracks = size(assignments, 1);
    for i = 1:numAssignedTracks
        trackIdx = assignments(i, 1);
        detectionIdx = assignments(i, 2);
        centroid = centroids(detectionIdx, :);
        bbox = bboxes(detectionIdx, :);
        correct(tracks(trackIdx).kalmanFilter, centroid);
        tracks(trackIdx).bbox = bbox;
        tracks(trackIdx).age = tracks(trackIdx).age + 1;
        tracks(trackIdx).totalVisibleCount = ...
            tracks(trackIdx).totalVisibleCount + 1;
        tracks(trackIdx).consecutiveInvisibleCount = 0; % Corrected field name here
    end
end

%% Update Unassigned Tracks
    function updateUnassignedTracks()
        for i = 1:length(unassignedTracks)
            ind = unassignedTracks(i);
                % Update track's age.
                tracks(ind).age = tracks(ind).age + 1;
                tracks(ind).consecutiveInvisibleCount = ...
                tracks(ind).consecutiveInvisibleCount + 1;
        end
    end

%% Delete Lost Tracks
function deleteLostTracks()
    if isempty(tracks)
        return;
    end

    invisibleForTooLong = 20; % Threshold for considering a track lost
    ageThreshold = 8; % Minimum length of track to be considered for deletion

    % Compute the fraction of the track's age for which it was visible.
    ages = [tracks(:).age];
    totalVisibleCounts = [tracks(:).totalVisibleCount];
    visibility = totalVisibleCounts ./ ages;

    % Find the indices of 'lost' tracks.
    lostInds = (ages < ageThreshold & visibility < 0.6) | ...
        [tracks(:).consecutiveInvisibleCount] >= invisibleForTooLong;

    % Delete lost tracks.
    tracks = tracks(~lostInds);
end

%% Create New Tracks
function createNewTracks()
    centroids = centroids(unassignedDetections, :);
    bboxes = bboxes(unassignedDetections, :);
    
    for i = 1:size(centroids, 1)
        centroid = centroids(i,:);
        bbox = bboxes(i, :);

        % Here, you need to provide two elements for InitialEstimateError, one for the position
        % and one for the velocity because you are using the 'ConstantVelocity' model.
        kalmanFilter = configureKalmanFilter('ConstantVelocity', ...
            centroid, [200, 50], [100, 25], 100);  % Tune these parameters as needed

        % Create a new track.
        newTrack = struct(...
            'id', nextId, ...
            'bbox', bbox, ...
            'kalmanFilter', kalmanFilter, ...
            'age', 1, ...
            'totalVisibleCount', 1, ...
            'consecutiveInvisibleCount', 0);

        % Add it to the array of tracks.
        tracks(end + 1) = newTrack;

        % Increment the next ID.
        nextId = nextId + 1;
    end
end

%% Display Tracking Results
function displayTrackingResults()
    % Convert the frame and mask to uint8 RGB.
    frame = im2uint8(frame);
    mask = uint8(repmat(mask, [1, 1, 3])) .* 255;
    
    minVisibleCount = 8;
    if ~isempty(tracks)
        % Noisy detections tend to result in short-lived tracks.
        % Only display tracks that have been visible for more than
        % a minimum number of frames.
        reliableTrackInds = ...
            [tracks(:).totalVisibleCount] > minVisibleCount;
        reliableTracks = tracks(reliableTrackInds);
        
        % Display the objects. If an object has not been detected
        % in this frame, display its predicted bounding box.
        if ~isempty(reliableTracks)
            bboxes = cat(1, reliableTracks.bbox);
            ids = int32([reliableTracks(:).id]);
            
            labels = cellstr(int2str(ids'));
            predictedTrackInds = ...
                [reliableTracks(:).consecutiveInvisibleCount] > 0;
            isPredicted = cell(size(labels));
            isPredicted(predictedTrackInds) = {' predicted'};
            labels = strcat(labels, isPredicted);
            
            frame = insertObjectAnnotation(frame, 'rectangle', ...
                bboxes, labels);
            mask = insertObjectAnnotation(mask, 'rectangle', ...
                bboxes, labels);
        end
    end
    
    % Draw on the mask.
    obj.maskPlayer.step(mask);
    
    % Draw on the frame.
    obj.videoPlayer.step(frame);
    
    % Write the frame to the video.
    videoWriter.step(frame);
end

end
